/*<html><pre>  -<a                             href="ttcnf.htm"
    >-------------------------------</a><a name="TOP">-</a>

    ttcnf.c
        Write comma-separated statistics for tt-CNF of 1 to 5 variables.

    The MASM32 assembler must be at least version 8.2.  The version bundled
    with DevStudio 2005 is too old.

    Copyright (c) 2005-2014 by C. Bradford Barber
    $Id: //product/ttcnf/main/src/ttcnf.c#2 $
    $DateTime: 2011/07/18 00:07:22 $$Change: 1403 $$Author: bbarber $
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <setjmp.h>
#include <errno.h>
#include <time.h>

#ifdef _WIN32           /* for VirtualAlloc() */
#include <windows.h>
#endif

#if __cplusplus
extern "C" {
  int isatty (int);
}

#elif _MSC_VER
#include <io.h>
#define isatty _isatty

#else
int isatty (int);  /* returns 1 if stdin is a tty
                   if "Undefined symbol" this can be deleted along with call in main() */
#endif

/*-<a                             href="qh-qhull.htm#TOC"
  >-------------------------------</a><a name="prompt">-</a>

  tt_prompt
    long prompt for ttcnf

  notes:
    do not use ',' since they become new columns
*/

char tt_prompt[]= "\n\
ttcnf - compute CNF expressions of 1 to 5 variables\n\
    http://www.qhull.org/ttcnf  %s\n\
\n\
Command line options:\n\
-M  Use MEM_LARGE_PAGES on Windows (1/3 slower in practice)\n\
\n\
Input options (stdin with one line per run):\n\
    # comment to end of line\n\
    kcnf 0x0405 -- run 4-CNF on 5 variables.  Use 'kcnf 0' for bucket size\n\
    steps -- show each step after adding one clause to the expression\n\
    maxstep n -- stop after step n\n\
    output-tt -- output truthtables as tt,truthtable\n\
\n\
Options to set initial truth tables -- may be on multiple lines\n\
    keep -- keep previous truth tables\n\
    start N N N ... -- set initial truth table (e.g. start 0xFFFF)\n\
    from N to N by N -- initialize range of truth tables \n\
";
char tt_prompt2[]= "\n\
Options to assign truth tables to one of 32 buckets\n\
    highlow -- record high-low for each nibble  (e.g. 0x1F is all high).\n\
       A high nibble is 6 9 or it has three bits\n\
    highlowbit -- record high bit for highlow==0\n\
    lownibble -- count non-zero low nibbles; exclude if high\n\
    lownibble2 -- count non-zero low nibbles; exclude if high;\n\
                  and test nibbles of rotated truth tables\n\
    nibble N -- nibble ending at bit N\n\
    nobuckets -- do not count buckets\n\
    numbit -- count number of 1 bits in truth table\n\
\n\
Output as comma-separated file on stdout\n\
    Run -- a run is initiated by kcnf N\n\
    Step -- a step corresponds to adding one clause to a k-CNF expression\n\
    kCNF -- the size of a clause (k) and number of variables (n) as bytes\n\
       If n > 5 then a clause may be fewer than k variables\n\
       For example 'kcnf 0x0406' is all clauses of 3 and 4 variables\n\
    StartTT -- highest truth table set by 'start N' or other options\n\
    MaxActive -- highest truth table that was new in the previous step, and made a difference\n\
    Seconds -- seconds since the start of the program\n\
    Stat Time -- seconds to compute the bucket statistics\n\
    Non-SAT -- total number of k-CNF expressions with truth table == 0\n\
    Tot TT -- total number of truth tables\n\
    Tot NoTT -- total number of unrepresented truth tables\n\
    Tot New -- total number of newly represented truth tables\n\
    NoTT-N -- bucket N as defined by the options\n\
       For 'nibble 0' NoTT-0 is all truth tables with low-order nibble == 0\n\
       number of unrepresented truth tables\n\
       new -- newly represented truth tables\n\
       seen -- previously represented truth tables\n\
       twice -- truth tables generated by two or more different lengths\n\
\n\
Example\n\
    echo \"kcnf 0x0305  start 0xffffffff nobuckets steps output-tt\" | ttcnf >ttcnf-0x0305-cnt.txt\n\
";

/* Symbols for tt_cnf.c.

Notes:
    Must match tt_all.sym
*/
#define   tt_FLAGS 0x03
#define     tt_NEW 0x01
#define     tt_SEEN 0x02        /* tt_statistics assumes tt_SEEN>>1==tt_NEW */
#define     tt_SEENtwice 0x03
#define     tt_NEW32  0x55555555
#define     tt_SEEN32 0xAAAAAAAA
#define   tt_NONE -1
#define   tt_NIBBLESperWord 8
#define   tt_NIBBLEmask 0xF
#define   tt_BITSperNibble 4
#define   tt_BITSperFlags 2
#define   tt_FLAGSperByte 4
#define   tt_FLAGSperInt 16
#define   tt_HIGHhalfOffset (0x80000000 >> 2)  /* Byte offset to bit 0x80000000 */
#define   tt_MAXline 255

#define tt_NUMbuckets 32
#define tt_MAXnibble 27
#define tt_IShighLow 0x200
#define tt_IShighLowBit 0x201
#define tt_ISlowNibble 0x202
#define tt_ISlowNibble2 0x203
#define tt_ISnumBit 0x204
#define tt_ISnoBuckets 0x205

#define tt_TRUE 1
#define tt_FALSE 0
#define tt_NULL 0
#define tt_1K 0x0400
#define tt_MEMhugeLocked (4 * tt_1K * tt_1K)   /* in Windows, use VirtualAlloc() with MEM_LARGE_PAGE */
#define tt_MAXstep 200  /* prevent infinite loops */

#define tt_ERRnone 0
#define tt_ERRinit 1
#define tt_ERRrun 2

#define boolT unsigned int
#define ttTruthtableT unsigned int

typedef struct ttStatsT ttStatsT;
typedef struct ttGlobalT ttGlobalT;
typedef struct ttTaskT ttTaskT;

struct ttStatsT {                   /* matches stStatsT in tt_all.sym */
    unsigned int totalNone;         /* Truth tables not generated so far */
    unsigned int totalNew;          /* Newly generated truth tables at this step */
    unsigned int totalTT;           /* Truth tables generated */
    unsigned int new[tt_NUMbuckets];
    unsigned int twice[tt_NUMbuckets];
    unsigned int seen[tt_NUMbuckets];
    unsigned int none[tt_NUMbuckets];
};

struct ttGlobalT {
        ttTaskT *curTask;           /* Current task for a tt_runTT, also in all_tt.sym */
        int initLength_K;
        int bitsPerTT;
        int maxLength;
        ttTruthtableT *truthtables;   /* initially zeros */
        ttTaskT *taskList;
        ttTaskT *lastTask;
        int curRun;
        unsigned int maxStart;
        unsigned int prevMaxStart;
        int wasRun;
        jmp_buf errexit;
        boolT NOerrexit:1;
        boolT ISlargeMem:1;
        boolT wasError:1;
        boolT isBuckets:1;
        time_t startTime;
        time_t statTime;   /* previous start of statistics */
        double totalStatTime;
};

struct ttTaskT {
                                    /* used by ttcnf.asm, offsets must match tt_all.sym */
        int kcnf;
        unsigned int maxStart;
        unsigned int totNonSAT;
        unsigned int maxActive;
        unsigned int totSame;       /* Generated tables same as previous at this step */
        unsigned int totSame2;      /*    High Dword */
        unsigned int bucketType;
        ttStatsT *stats;
        char line[tt_MAXline+1];
                                    /* not used by tt_all.sym  */
        char *startValues;
        int linenum;
        int step;
        int run;
        int maxStep;
        boolT isSteps:1;
        boolT isKeep:1;
        boolT outputTT:1;
        unsigned int fromTT, toTT, byTT;   /* if set, toTT and byTT are not 0  */
        ttTaskT *next;
};

char *tt_version ="2014.1 2014/1/26";

ttGlobalT *tt_global;

unsigned int newBits[] = {
    0x00000001, 0x00000004, 0x00000010, 0x00000040,
    0x00000100, 0x00000400, 0x00001000, 0x00004000,
    0x00010000, 0x00040000, 0x00100000, 0x00400000,
    0x01000000, 0x04000000, 0x10000000, 0x40000000
};
unsigned char tt_highNibbles[] = {
    0,0,0,0,
    0,0,1,1,   /* 0x6,0x7 */
    0,1,0,1,   /* 0x9,0xb */
    0,1,1,0    /* 0xd,0xe */
};

/* Bitcount() -- Not used
   http://gurmeet.net/puzzles/fast-bit-counting-routines/
*/
#define TWO(c)     (0x1u << (c))
#define MASK(c) \
    (((unsigned int)(-1)) / (TWO(TWO(c)) + 1u))
#define COUNT(x,c) \
    ((x) & MASK(c)) + (((x) >> (TWO(c))) & MASK(c))
int bitcount (unsigned int n)  {
    n = COUNT(n, 0) ;
    n = COUNT(n, 1) ;
    n = COUNT(n, 2) ;
    n = COUNT(n, 3) ;
    n = COUNT(n, 4) ;
    /* n = COUNT(n, 5) ;    for 64-bit integers */
    return n ;
}

void tt_adjustTokenPrivileges(char* TokenPrivilege);
void *tt_allocZero(char *what, int size);
void tt_errexit(int errType, char *errmsg, ...);
void tt_freeGlobal(ttGlobalT **tt_globalp);
int tt_getBucket(ttTaskT *task, unsigned int truthtable);
void tt_initGlobal(ttGlobalT **tt_globalp);
void tt_initTruthtables(ttGlobalT *tt_global);
void tt_keepSeenTT(ttTruthtableT *truthtables, unsigned int maxStart);
void tt_maximizeUnsigned(unsigned int *variablep, unsigned int value);
ttTaskT *tt_newTask(ttGlobalT *tt_global);
void tt_outputTT(ttTaskT *task, ttTruthtableT *truthtables);
void tt_readArgs(ttGlobalT *tt_global, int argc, char *argv[]);
int tt_readNumber(char **stringp, char *where);
void tt_readTasks(ttGlobalT *tt_global);
unsigned int tt_readUnsigned(char **stringp, char *where);
boolT tt_readWord(char **stringp, char *word);
void tt_runTasks(ttGlobalT *tt_global);
void tt_runTT(ttTaskT *task, ttTruthtableT truthtables[]); /* for testing */
void tt_skipWord(char **stringp);
void tt_statistics(ttTaskT *task, ttTruthtableT *truthtables);
boolT tt_updateSeenTT(ttTruthtableT *truthtables, unsigned int maxStart);
void tt_writeStats(ttTaskT *task);

void tt_runTT_c(ttTaskT *task, ttTruthtableT truthtables[]);
boolT tt_setTT_c(ttTruthtableT *truthtables, unsigned int tt);
void tt_statisticsTT_c(ttTaskT *task, ttTruthtableT truthtables[]);
int tt_highbit_c(unsigned int tt);

#ifdef _WIN32
void tt_strErrorText( char* errText, int errTextMax, DWORD errCode);
#endif


/*-<a                             href="qh-qhull.htm#TOC"
  >-------------------------------</a><a name="main">-</a>

  main( argc, argv )
    processes the command line, calls qhull() to do the work, and exits

  design:
    initializes data structures
    writes the output
    frees memory
*/
int main(int argc, char *argv[]) {
    int exitcode;

    if ((argc == 1) && isatty( 0 /*stdin*/)) {
        fprintf(stdout, tt_prompt, tt_version);
        fprintf(stdout, tt_prompt2);
        exit(tt_ERRnone);
    }
    tt_initGlobal(&tt_global);
    tt_global->NOerrexit= tt_TRUE;  /* no more setjmp */
    tt_readArgs(tt_global, argc, argv);
    exitcode= setjmp(tt_global->errexit);
    tt_global->NOerrexit= tt_FALSE;
    if (!exitcode) {
        tt_readTasks(tt_global);
        tt_initTruthtables(tt_global);
        tt_runTasks(tt_global);
    }
    tt_global->NOerrexit= tt_TRUE;  /* no more setjmp */
    tt_freeGlobal(&tt_global);
    return exitcode;
}/*main*/


/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_freeGlobal">-</a>

    tt_freeGlobal(tt_global)
        Free global data structure and all memory

    notes:
*/
void tt_freeGlobal(ttGlobalT **tt_globalp) {
    ttGlobalT *tt_global= *tt_globalp;
    if (tt_global) {
        ttTaskT *task= tt_global->taskList;

        if (tt_global->truthtables) {
#ifdef _WIN32
            if (tt_global->maxLength >= tt_MEMhugeLocked)
                VirtualFree(tt_global->truthtables, tt_NULL, MEM_RELEASE);
            else
                free(tt_global->truthtables);
#else
            free(tt_global->truthtables);
#endif
        }
        while (task) {
            ttTaskT *nextTask= task->next;
            free(task);
            task= nextTask;
        }
        free(tt_global);
        *tt_globalp= tt_NULL;
    }
}/*tt_freeGlobal*/


/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_getBucket">-</a>

    tt_getBucket( task, truthtable )
        Return bucket (0..15) for truthtable
    example:
        int bucket= tt_getBucket(task, truthtable);
    notes:
        For statistics
*/
int tt_getBucket(ttTaskT *task, unsigned int truthtable) {
    int bucket= 0;
    int i;

    if (task->bucketType == tt_NONE) {
        return 0;                       /* already reported as error */
    }else if (task->bucketType <= tt_MAXnibble) {
        bucket= (truthtable >> task->bucketType) & tt_NIBBLEmask;
    }else if (task->bucketType == tt_IShighLow) {
        unsigned int tt32= truthtable;
        for (i= tt_NIBBLESperWord; i--; ) {
            int nibble= tt32 & tt_NIBBLEmask;
            int isHigh= tt_highNibbles[nibble];
            bucket |= (isHigh << tt_NIBBLESperWord);
            bucket >>= 1;
            tt32 >>= tt_BITSperNibble;
        }
    }else if (task->bucketType == tt_IShighLowBit) {
        unsigned int tt32= truthtable;
        for (i= tt_NIBBLESperWord; i--; ) {
            int nibble= tt32 & tt_NIBBLEmask;
            int isHigh= tt_highNibbles[nibble];
            tt32 >>= tt_BITSperNibble;
            if (isHigh)
                break;
        }
        if (i>=0)
            bucket= tt_NUMbuckets;
        else
            bucket= tt_highbit_c(truthtable);
    }
    if (bucket >= tt_NUMbuckets)
        bucket= tt_NUMbuckets-1;
    return bucket;
} /*tt_getBucket*/


/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_initGlobal">-</a>

    tt_initGlobal(tt_global)
        Initialize global data structure


    notes:
*/
void tt_initGlobal(ttGlobalT **tt_globalp) {
    ttGlobalT *tt_global= (ttGlobalT*)calloc(1, sizeof(ttGlobalT));
    if (!tt_global) {
        fprintf(stderr, "No memory available to run program.\n");
        exit(1);
    }
    *tt_globalp= tt_global;
    tt_global->bitsPerTT= 2;
    tt_global->initLength_K= 64;
    tt_global->totalStatTime= 0.0;
} /*tt_initGlobal*/


/*-<a                             href="ttcnf.htm#TOC"
  >-------------------------------</a><a name="tt_initTruthtables">-</a>

  tt_initTruthtables(tt_global)
        Allocate truthtables for tt_global->maxStart

  returns:
    TaskList

  notes:
*/
void tt_initTruthtables(ttGlobalT *tt_global) {
    double byteCountR= ((double)tt_global->maxStart + tt_FLAGSperByte - 1)/tt_FLAGSperByte;
    int byteLength= (int)byteCountR;
    tt_global->truthtables= (ttTruthtableT*)tt_allocZero("truthtables", byteLength);
    tt_global->maxLength= byteLength;
}

/*-<a                             href="ttcnf.htm#TOC"
>-------------------------------</a><a name="tt_allocZero">-</a>

    tt_allocZero( what, size)
        Alloccate and zero size bytes of memory
    returns:
        memory, report error if none
    example:
        task= (ttTaskT)tt_allocZero("CNF Task", sizeof(ttTaskT));
    notes:
        May need MEM_LARGE_PAGES, if so also need SE_LOCK_MEMORY_PRIVILEGE from AdjustTokenPrivileges()
            SIZE_T largePageMin= GetLargePageMinimum(void);
            int cntLargePages= (size + largePageMin -1) / largePageMin;
            int largeSize= cntLargePages * largePageMin;
            object= VirtualAlloc(tt_NULL, largeSize, MEM_COMMIT+MEM_LARGE_PAGES, PAGE_READWRITE);
        PAGE_READWRITE requires special priveleges
*/
void *tt_allocZero(char *what, int size) {
    void *object;

#ifdef _WIN32
    if (size >= tt_MEMhugeLocked) {
        int flags= MEM_COMMIT;
        if (tt_global->ISlargeMem)
            flags |= MEM_LARGE_PAGES;
        object= VirtualAlloc(tt_NULL, size,  flags, PAGE_EXECUTE_READWRITE);
        if (!object) {
            DWORD errCode= GetLastError();
            char errText[100];
            tt_strErrorText( errText, sizeof(errText), errCode);
            tt_errexit(tt_ERRinit, "Insufficient memory to allocate %s of %d bytes.  Error %d, %s",
                                what, size, errCode, errText);
        }
    }else {
        object= calloc(size, 1);
    }
#else
    object= calloc(size, 1);
#endif
    if (!object) {
        tt_errexit(tt_ERRinit, "Insufficient memory to allocate %s of %d bytes", what, size);
    }
    return object;
}/*tt_allocZero*/

/*-<a                             href="ttcnf.htm#TOC"
>-------------------------------</a><a name="tt_errexit">-</a>

    tt_errexit( errType, message, args)
        Exit program and report error
        tt_global is the global data array
    returns:
        never returns
    example:
        tt_errexit(tt_ERRinit, "Insufficient memory to allocate %s of %d bytes", what, size);
    notes:
        See tt_errlog()
        See K&R Edition 2, p. 174, error()
*/
void tt_errexit(int errType, char *errmsg, ...) {
    va_list args;

    va_start(args, errmsg);
    fprintf(stderr, "Error ttcnf: ");
    vfprintf(stderr, errmsg, args);
    fprintf(stderr, "\n");
    va_end(args);

    if (tt_global->NOerrexit)
        exit(1);
    longjmp(tt_global->errexit, errType);  /* never returns */
}/*tt_errexit*/

/*-<a                             href="ttcnf.htm#TOC"
>-------------------------------</a><a name="tt_errlog">-</a>

    tt_errlog( errType, message, args)
        Log error and continue
    returns:
        sets tt_global->wasError
    example:
        tt_errlog(tt_ERRinit, "Insufficient memory to allocate %s of %d bytes", what, size);
    notes:
        See tt_errexit()
        See K&R Edition 2, p. 174, error()
*/
void tt_errlog(int errType, char *errmsg, ...) {
    va_list args;

    tt_global->wasError= tt_TRUE;
    va_start(args, errmsg);
    fprintf(stderr, "Error ttcnf: ");
    vfprintf(stderr, errmsg, args);
    fprintf(stderr, "\n");
    va_end(args);
}/*tt_errlog*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_maximizeUnsigned">-</a>

    tt_maximizeUnsigned(variable, value)
        Maximize variable as value
    notes:
*/
void tt_maximizeUnsigned(unsigned int *variablep, unsigned int value) {
    if (value > *variablep)
        *variablep= value;
}/*tt_maximizeUnsigned*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_newTask">-</a>

    tt_newTask(tt_global)
        Allocate new task

    returns:
        Task

    notes:
*/
ttTaskT *tt_newTask(ttGlobalT *tt_global) {
    ttTaskT *task;
    task= (ttTaskT*)tt_allocZero("CNF Task", sizeof(ttTaskT));
    if (tt_global->lastTask)
        tt_global->lastTask->next= task;
    else
        tt_global->taskList= task;
    tt_global->lastTask= task;
    task->bucketType= tt_NONE;
    task->kcnf= tt_NONE;
    task->run= tt_global->curRun;
    return task;
}/*tt_newTask*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_readTasks">-</a>

    tt_readTasks(tt_global)
        Read task list from stdin

    returns:
        TaskList

    notes:
*/
void tt_readTasks(ttGlobalT *tt_global) {
    char *s, line[tt_MAXline+1];
    int linecount= 0;
    ttTaskT *task;
    boolT isHeader= tt_TRUE;

    tt_global->curRun= 1;
    while (fgets(line, tt_MAXline+1, stdin)) { /* for each input line */
        linecount++;
        if (strlen(line) == tt_MAXline) {
            tt_errlog(tt_ERRinit, "Line %d is longer than %d bytes\n%s", linecount, tt_MAXline-2, line);
            continue;
        }
        if (isHeader && *line == '#') {
            fprintf(stdout, "%s", line);
            continue;
        }
        s= line;
        while (isspace(*s))
            s++;
        if (!*s)
            continue;
        isHeader= tt_FALSE;
        task= tt_newTask(tt_global);
        task->linenum= linecount;
        strcat(task->line, line);
        s= task->line;
        while (isspace(*s))
            s++;
        while (*s) {
            if (*s == '#')
                break;
            if (tt_readWord(&s, "kcnf")) {
                task->kcnf= tt_readNumber(&s, "kcnf");
                tt_global->curRun++;
            }else if (tt_readWord(&s, "steps")) {
                task->isSteps= tt_TRUE;
            }else if (tt_readWord(&s, "keep")) {
                if (!tt_global->prevMaxStart)
                    tt_errlog(tt_ERRinit, "Define a task using 'start N' prior to 'keep': %s", task->line);
                task->isKeep= tt_TRUE;
                task->maxStart= tt_global->prevMaxStart;
            }else if (tt_readWord(&s, "output-tt")) {
                task->outputTT= tt_TRUE;
            }else if (tt_readWord(&s, "highlow")) {
                if (task->bucketType != tt_NONE)
                    tt_errlog(tt_ERRinit, "Already defined a bucket type prior to 'highlow': %s", task->line);
                task->bucketType= tt_IShighLow;
                tt_global->isBuckets= tt_TRUE;
            }else if (tt_readWord(&s, "highlowbit")) {
                if (task->bucketType != tt_NONE)
                    tt_errlog(tt_ERRinit, "Already defined a bucket type prior to 'highlowbit': %s", task->line);
                task->bucketType= tt_IShighLowBit;
                tt_global->isBuckets= tt_TRUE;
            }else if (tt_readWord(&s, "lownibble")) {
                if (task->bucketType != tt_NONE)
                    tt_errlog(tt_ERRinit, "Already defined a bucket type prior to 'lownibble': %s", task->line);
                task->bucketType= tt_ISlowNibble;
                tt_global->isBuckets= tt_TRUE;
            }else if (tt_readWord(&s, "lownibble2")) {
                if (task->bucketType != tt_NONE)
                    tt_errlog(tt_ERRinit, "Already defined a bucket type prior to 'lownibble2': %s", task->line);
                task->bucketType= tt_ISlowNibble2;
                tt_global->isBuckets= tt_TRUE;
            }else if (tt_readWord(&s, "nobuckets")) {
                if (task->bucketType != tt_NONE)
                    tt_errlog(tt_ERRinit, "Already defined a bucket type prior to 'nobuckets': %s", task->line);
                task->bucketType= tt_ISnoBuckets;
            }else if (tt_readWord(&s, "numbit")) {
                if (task->bucketType != tt_NONE)
                    tt_errlog(tt_ERRinit, "Already defined a bucket type prior to 'numbit': %s", task->line);
                task->bucketType= tt_ISnumBit;
                tt_global->isBuckets= tt_TRUE;
            }else if (tt_readWord(&s, "nibble")) {
                if (task->bucketType != tt_NONE)
                    tt_errlog(tt_ERRinit, "Already defined a bucket type prior to 'nibble': %s", task->line);
                task->bucketType= tt_readNumber(&s, "nibble");
                if (task->bucketType > tt_MAXnibble)
                    tt_errlog(tt_ERRinit, "Max bit offset for 'nibble' is %d", tt_MAXnibble);
                tt_global->isBuckets= tt_TRUE;
            }else if (tt_readWord(&s, "maxstep")) {
                task->maxStep= tt_readNumber(&s, "maxstep");
            }else if (tt_readWord(&s, "start")) {
                if (task->startValues)
                    tt_errlog(tt_ERRinit, "More than one 'start' per line followed by one or more numbers: %s", task->line);
                task->startValues= s;
                if (!isdigit(*s))
                    tt_errlog(tt_ERRinit, "Missing number after 'start' for truthtable: %s", task->line);
                while (isdigit(*s)) {
                    unsigned int start= tt_readUnsigned(&s, "start");
                    tt_maximizeUnsigned(&task->maxStart, start);
                }
            }else if (tt_readWord(&s, "from")) {
                task->fromTT= tt_readUnsigned(&s, "from");
                if (!tt_readWord(&s, "to")) {
                    tt_errlog(tt_ERRinit, "Missing 'to N' after 'from N' at '%s'", *s);
                    continue;
                }
                task->toTT= tt_readUnsigned(&s, "to");
                tt_maximizeUnsigned(&task->maxStart, task->toTT);
                if (!tt_readWord(&s, "by")) {
                    tt_errlog(tt_ERRinit, "Missing 'by N' after 'to N' at '%s'", *s);
                    continue;
                }
                task->byTT= tt_readUnsigned(&s, "by");
                if (!task->byTT || !task->toTT)
                    tt_errlog(tt_ERRinit, "Either 'by N' or 'to N' is zero '%s'", task->line);
            }else if (isalnum(*s)) {
                tt_errlog(tt_ERRinit, "Unknown keyword at '%s'", s);
                tt_skipWord(&s);
            }else {
                tt_errlog(tt_ERRinit, "Unknown notation or character (%d) at '%s'", *s, s);
                tt_skipWord(&s);
            }
        }/*each word*/
        tt_maximizeUnsigned(&tt_global->maxStart, task->maxStart);
        tt_global->prevMaxStart= task->maxStart;
        if (task->kcnf != tt_NONE && task->bucketType == tt_NONE)
            tt_errlog(tt_ERRinit, "Missing bucket type.  Use 'nibble N', 'highlow', 'lownibble', etc.: %s", task->line);
    }/*each line*/
    if (tt_global->wasError)
        tt_errexit(tt_ERRinit, "Input errors.  Run 'ttcnf' for help.");
    if (!tt_global->maxStart)
        tt_errexit(tt_ERRinit, "No 'start N' values defined (non-zero)");
}/*tt_readTasks*/


/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_readNumber">-</a>

    tt_readNumber( string, where)
        Returns number at string
    returns:
        int
        if error, reports where
*/
int tt_readNumber(char **stringp, char *where) {
    char *endp;
    long value;

    errno= 0;
    if (!isdigit(**stringp) && **stringp != '-')
        tt_errexit(tt_ERRinit, "Expecting a number for %s.  Instead got '%s'", where, *stringp);
    if (tolower((*stringp)[1]) == 'x')
        value= strtol(*stringp, &endp, 16);
    else
        value= strtol(*stringp, &endp, 10);
    if (errno)
        tt_errexit(tt_ERRinit, "Invalid number for %s at '%s'", where, *stringp, strerror(errno));
    while (isspace(*endp))
        endp++;
    *stringp= endp;
    return (int)value;
}/*tt_readNumber*/


/*-<a                             href="ttcnf.htm#TOC"
>-------------------------------</a><a name="tt_readUnsigned">-</a>

    tt_readUnsigned( string, where)
        Returns number at string
    returns:
        unsigned int
        if error, reports where
        otherwise, errno is 0
*/
unsigned int tt_readUnsigned(char **stringp, char *where) {
    char *endp;
    unsigned long value;

    errno= 0;
    if (!isdigit(**stringp))
        tt_errexit(tt_ERRinit, "Expecting an unsigned number for '%s'.  Instead got '%s'", where, *stringp);
    if (tolower((*stringp)[1]) == 'x')
        value= strtoul(*stringp, &endp, 16);
    else
        value= strtoul(*stringp, &endp, 10);
    if (errno)
        tt_errexit(tt_ERRinit, "Invalid number for %s at '%s'", where, *stringp, strerror(errno));
    while (isspace(*endp))
        endp++;
    *stringp= endp;
    return (unsigned int)value;
}/*tt_readUnsigned*/


/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_readWord">-</a>

    tt_readWord( string, "word")
        Returns true if next word is "word"
    returns:
        if true, advances string after the word and trailing spacing
    notes:
        case sensitive
*/
boolT tt_readWord(char **stringp, char *word) {
    size_t wordLen= strlen(word);

    if (0==strncmp(*stringp, word, wordLen)) {
        if (!isalnum((*stringp)[wordLen])) {
            (*stringp) += wordLen;
            while (isspace(**stringp))
                (*stringp)++;
            return tt_TRUE;
        }
    }
    return tt_FALSE;
}/*tt_readWord*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_skipWord">-</a>

    tt_skipWord( string)
        Advances string after next word or char
    notes:
        Skips spacing after word
*/
void tt_skipWord(char **stringp) {
    char c= **stringp;
    if (isalnum(c) || c =='_') {
        while (tt_TRUE) {
            (*stringp)++;
            c= **stringp;
            if (c != '_' && !isalnum(c))
                break;
        }
    }else {
        (*stringp)++;
    }
    while (isspace(**stringp))
        (*stringp)++;
}/*tt_skipWord*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_runTasks">-</a>

    tt_runTasks(tt_global)
        Run each task

    notes:
*/
void tt_runTasks(ttGlobalT *tt_global) {
    char curTime[40];
    char *s;
    time_t now;
    unsigned int start;
    ttTaskT *task;

    time(&now);
    tt_global->startTime= now;
    strftime(curTime, sizeof(curTime), "%Y/%m/%d %H:%M", localtime(&now));
    fprintf(stdout, "# Results of 'ttcnf' on %s. \n\n", curTime);
    tt_writeStats(tt_NULL);
    fprintf(stdout, "\n");

    for (task= tt_global->taskList; task; task= task->next) {   /* foreach task */
        if (tt_global->wasRun) {
            tt_global->wasRun= tt_FALSE;
            fprintf(stdout, "\n");
            if (task->isKeep) {
                tt_keepSeenTT(tt_global->truthtables, task->maxStart);
            }else {
                memset(tt_global->truthtables, 0, tt_global->maxLength);
            }
        }
        if (task->maxStart) {
            if ((s= task->startValues)) {
                while (isdigit(*s)) {
                    start= tt_readUnsigned(&s, "start");
                    tt_setTT_c(tt_global->truthtables, start);
                }
            }
            if (task->toTT) {
                for (start= task->fromTT; start <= task->toTT; start += task->byTT)
                    tt_setTT_c(tt_global->truthtables, start);
            }
        }/*maxStart*/
        fprintf(stdout, "%s", task->line);
        if (task->kcnf == tt_NONE)
            continue;
        tt_global->curTask= task;   /* for ztt4cnf0.asm */
        tt_global->wasRun= tt_TRUE;
        task->step=0;   /* foreach step */
        while (tt_TRUE) {
            task->step++;
            task->maxActive= 0;
            if (task->step > tt_MAXstep)
                tt_errexit(tt_ERRrun, "More than %d steps.  Infinite loop?", tt_MAXstep);
            if (task->maxStep && task->step > task->maxStep)
                break;
            if (task->kcnf)
                tt_runTT_c(task, tt_global->truthtables);
            if (task->isSteps) {
                tt_statistics(task, tt_global->truthtables);
                tt_writeStats(task);
                task->totSame= 0;
                task->totSame2= 0;
            }
            if (!task->maxActive)
                break;
        }/*foreach step*/
        if (!task->isSteps) {
            tt_statistics(task, tt_global->truthtables);
            tt_writeStats(task);
        }
        if (task->outputTT)
            tt_outputTT(task, tt_global->truthtables);
    }/*foreach task*/
    fprintf(stdout, "\nTotal Statistics Time = %0.0f seconds\n", tt_global->totalStatTime);
    fprintf(stdout, tt_prompt, tt_version);
    fprintf(stdout, tt_prompt2);
}/*tt_runTasks*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_runTT">-</a>

    tt_runTT( task, truthtables )

    notes:
*/
void tt_runTT(ttTaskT *task, ttTruthtableT *truthtables) {
    /* will be asm program, for testing */
    tt_setTT_c(truthtables, task->step);
}/*tt_runTT*/


/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_statistics">-</a>

    tt_statistics( task, truthtables )
    notes:
*/
void tt_statistics(ttTaskT *task, ttTruthtableT truthtables[]) {
    ttStatsT *stats;
    int i;

    tt_global->statTime= time(tt_NULL);
    stats= (ttStatsT*)tt_allocZero("Statistics", sizeof(ttStatsT));;
    if (task->stats)
        free(task->stats);
    task->stats= stats;

    if (task->bucketType != tt_ISnoBuckets) {
        tt_statisticsTT_c(task, truthtables);
        for (i=tt_NUMbuckets; i--; ) {
            stats->totalNone += stats->none[i];
            stats->totalNew += stats->new[i];
            stats->totalTT += stats->new[i] + stats->seen[i] + stats->twice[i];
        }
    }else {
        unsigned int truthtable= 0;
        ttTruthtableT *truthtablesp= truthtables;
        while (tt_TRUE) {
            unsigned int ttflags32= *truthtablesp++;
            for (i= tt_FLAGSperInt; i--; ) {
                int ttflags= ttflags32 & tt_FLAGS;
                if (ttflags) {
                    if (ttflags == tt_NEW)
                        stats->totalNew++;
                    stats->totalTT++;
                }else {
                    stats->totalNone++;
                }
                ttflags32 >>= tt_BITSperFlags;
                if (truthtable >= task->maxStart)
                    return;
                truthtable++;
            }
        }
    }

/******** equivalent C code for nibbleN and highlow *************
    while (tt_TRUE) {
        unsigned int ttflags32= *truthtablesp++;
        for (i= tt_FLAGSperInt; i--; ) {
            int ttflags= ttflags32 & tt_FLAGS;
            int bucket= tt_getBucket(task, truthtable);
            if (ttflags) {
                if (ttflags == tt_SEENtwice) {
                    stats->twice[bucket]++;
                }else if (ttflags == tt_NEW) {
                    stats->new[bucket]++;
                    stats->totalNew++;
                }else {
                    stats->seen[bucket]++;
                }
            }else {
                stats->none[bucket]++;
                stats->totalNone++;
            }
            ttflags32 >>= tt_BITSperFlags;
            if (truthtable >= task->maxStart)
                return;
            truthtable++;
        }
    }
***********************/
}/*tt_statistics*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_keepSeenTT">-</a>

    tt_keepSeenTT( truthtables )
        Convert seen truthtables to new
    notes:
        For 'keep' option
*/
void tt_keepSeenTT(ttTruthtableT *truthtables, unsigned int maxStart) {
    ttTruthtableT *truthtablesp= truthtables;
    int i;

    for (i= maxStart/tt_FLAGSperInt+1; i--; ) {   /* foreach truthtable */
        unsigned int ttflags32= *truthtablesp;
        unsigned int ttseen32= ttflags32 & tt_SEEN32;
        unsigned int ttnew32= ttseen32 >> 1;
        unsigned int ttflags32_new;
        if (ttseen32) {
            ttflags32_new= (ttflags32 | ttnew32) ^ ttseen32;
            *(truthtablesp++)= ttflags32_new;
        }else {
            truthtablesp++;
        }
    }
} /*tt_keepSeenTT*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_updateSeenTT">-</a>

    tt_updateSeenTT( truthtables )
    returns:
        Moves new truthtables, not previously seen, to seen
    notes:
        Not used.
*/
boolT tt_updateSeenTT(ttTruthtableT *truthtables, unsigned int maxStart) {
    ttTruthtableT *truthtablesp= truthtables;
    boolT wasNew= tt_FALSE;
    int i;

    for (i= maxStart/tt_FLAGSperInt+1; i--; ) {   /* foreach truthtable */
        unsigned int ttflags32= *truthtablesp;
        unsigned int ttseen32= ttflags32 & tt_SEEN32;
        unsigned int ttnew32= ttflags32 & tt_NEW32;
        unsigned int ttflags32_new;
        ttnew32 &= (~ttseen32) >> 1;
        if (ttnew32) {
            ttflags32_new= (ttflags32 ^ ttnew32) | (ttnew32 << 1);
            *(truthtablesp++)= ttflags32_new;
            wasNew= tt_TRUE;
        }else {
            truthtablesp++;
        }
    }
    return wasNew;
} /*tt_updateSeenTT*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_writeStats">-</a>

    tt_writeStats( task )
        Write statistics for task
        If task.isNull, write header
    notes:
*/

void tt_writeStats(ttTaskT *task) {
    int i;
    time_t now= time(tt_NULL);

    if (!task) {
        fprintf(stdout, "Run,Step,kCNF,StartTT,MaxActive,Seconds,Stat Time,Non-SAT,Tot TT,Tot NoTT,Tot Same,Tot New");
        if (tt_global->isBuckets) {
            for (i=0; i<tt_NUMbuckets; i++)
                fprintf(stdout, ",NoTT-%d", i);
        }
    }else {
        ttStatsT *stats= task->stats;
        double  statsTime;
        double  totSame= 1.0 * task->totSame + task->totSame2 * (4.0 * 0x40000000);

        if (!stats)
            return;                                     /* header line must match commas */
        statsTime= difftime(now, tt_global->statTime);
        tt_global->totalStatTime += statsTime;
        fprintf(stdout, "%d,%d,%0#.4x,%0#.8x,%0#.8x", task->run, task->step, task->kcnf, task->maxStart, task->maxActive);
        fprintf(stdout, ",%0.0f,%0.0f", difftime(now, tt_global->startTime), statsTime);
        fprintf(stdout, ",%u,%u,%u,%0.0f,%u", task->totNonSAT, stats->totalTT, stats->totalNone, totSame,stats->totalNew);
        if (task->bucketType != tt_ISnoBuckets) {
            for (i=0; i<tt_NUMbuckets; i++)
                fprintf(stdout, ",%u", stats->none[i]);
            if (task->isSteps || task->isKeep) {
                fprintf(stdout, "\n,,,,,,,,,,,new");            /* commas must match header line */
                for (i=0; i<tt_NUMbuckets; i++)
                    fprintf(stdout, ",%u", stats->new[i]);
                fprintf(stdout, "\n,,,,,,,,,,,seen");
                for (i=0; i<tt_NUMbuckets; i++)
                    fprintf(stdout, ",%u", stats->seen[i]);
                fprintf(stdout, "\n,,,,,,,,,,,twice");
                for (i=0; i<tt_NUMbuckets; i++)
                    fprintf(stdout, ",%u", stats->twice[i]);
            }
        }
    }
    fprintf(stdout, "\n");
    fflush(stdout);
}/*tt_writeStats*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_outputTT">-</a>

    tt_outputTT( task, truthtables )
    notes:
       The number of instances of each truth table is not useful.
       Once a truth table is generated, its count can be increased indefinitely by adding redundant clauses.
       The assembly language code is optimized to stop processing when new truthtables will not be generated.
*/
void tt_outputTT(ttTaskT *task, ttTruthtableT truthtables[]) {
    int i;

    unsigned int truthtable= 0;
    ttTruthtableT *truthtablesp= truthtables;
    while (tt_TRUE) {
        unsigned int ttflags32= *truthtablesp++;
        for (i= tt_FLAGSperInt; i--; ) {
            int ttflags= ttflags32 & tt_FLAGS;
            if (ttflags) {
                fprintf(stdout, "tt,%u\n", truthtable);
            }
            ttflags32 >>= tt_BITSperFlags;
            if (truthtable >= task->maxStart)
                return;
            truthtable++;
        }
    }
    fflush(stdout);
}/*tt_outputTT*/

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_readArgs">-</a>

    tt_readArgs( task )
        -M sets SE_LOCK_MEMORY_NAME for MEM_LARGE_PAGES
    notes:
*/

void tt_readArgs(ttGlobalT *tt_global, int argc, char *argv[]) {
    int i;

    for (i= 1; i < argc; i++) {
        if (0 == strcmp(argv[i], "-M")) {
#ifdef _WIN32
            tt_adjustTokenPrivileges(SE_LOCK_MEMORY_NAME);
#endif
            tt_global->ISlargeMem= tt_TRUE;
        }else {
            tt_errexit(tt_ERRinit, "Unknown input argument '%s'.  Run tt_cnf without args", argv[i]);
        }
    }
}/*tt_readArgs*/

#ifdef _WIN32
/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_DisplayErrorText">-</a>

    tt_DisplayErrorText( GetLastError() )
        Write error text to stdout
    notes:
        Search for netmsg.dll for networking errors (2100..2999)
*/

void tt_strErrorText( char* errText, int errTextMax, DWORD errCode) {

    if (!FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM, tt_NULL, errCode, tt_NULL,
                                errText, errTextMax, tt_NULL)) {
        strncpy(errText, "Unknown system error.", errTextMax);
    }
}

/*-<a                             href="ttcnf.htm#TOC"
    >-------------------------------</a><a name="tt_adjustTokenPrivileges">-</a>

    tt_adjustTokenPrivileges( privilege )
        turn on privilege for access token of current task
    example:
        tt_adjustTokenPrivileges(SE_LOCK_MEMORY_PRIVILEGE);
    notes:
        http://www.winterdom.com/dev/security/tokens.html
*/

void tt_adjustTokenPrivileges(char* TokenPrivilege) {
    HANDLE   hToken;     /* process token */
    TOKEN_PRIVILEGES tp;    /* token provileges */
    LUID     luid;
    HANDLE hProc= GetCurrentProcess();

    if (!OpenProcessToken(hProc, TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)) {
        DWORD errCode= GetLastError();
        char errText[100];
        tt_strErrorText( errText, sizeof(errText), errCode);
        tt_errexit(tt_ERRinit, "Not able to open process token.  Error %d, %s",
                            errCode, errText);
    }
    if (!LookupPrivilegeValue (NULL, TokenPrivilege, &luid)) {
        DWORD errCode= GetLastError();
        char errText[100];

        CloseHandle (hToken);
        tt_strErrorText( errText, sizeof(errText), errCode);
        tt_errexit(tt_ERRinit, "Not able to lookup %s.  Error %d, %s",
                            TokenPrivilege, errCode, errText);
    }
    ZeroMemory (&tp, sizeof (tp));
    tp.PrivilegeCount = 1;
    tp.Privileges[0].Luid = luid;
    tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
    if (!AdjustTokenPrivileges (hToken, FALSE, &tp, 0, NULL, NULL)) {
        DWORD errCode= GetLastError();
        char errText[100];

        CloseHandle (hToken);
        tt_strErrorText( errText, sizeof(errText), errCode);
        tt_errexit(tt_ERRinit, "Not able to adjust privilege %s.  Error %d, %s",
                            TokenPrivilege, errCode, errText);
    }
}/*tt_adjustTokenPrivileges*/

#endif
