;ztt4cnf0.lst
;Copyright (C) 2004-2005 by C. Bradford Barber
;$Id: //product/ttcnf/main/src/ztt4cnf0.asm#1 $
;$DateTime: 2008/03/16 14:47:31 $$Change: 844 $$Author: bbarber $

	include tt_all.sym

;;gather statistics for bitarray truth table CNF
;
;	EAX,EBX,ECX,EDX= masks for each variable
;	    EAX= 0 if < 4-CNF, etc.
;	    Choice of 0ff00ff00h, 0f0f0f0f0h, 0cccccccch, 0aaaaaaaah
;			ttcnf0 cycles through all the complements
;	ESI= previous TruthTable, unused bits are 0
;	EDI= Truthtables
;	EBP:highbits4 = complement mask -- max down to 0 -- e.g., 3000h for 2-CNF
;	    EBP:bit11= varE (00800h)
;			The fifth variable, E or 0FFFF0000h, is generated explicitly.
;	    EBP:bit10= bit was set (00400h)
;returns:
;	C set if ttwas0 (i.e., new or seen->twice)
;	EAX,EBX,ECX,ESI,EDI preserved
;	EBP.ttwas0 and ttsamemask updated (rest preserved)
;	set bits for Truthtables
;	    i is a truthtable 
;notes:
;	This is the inner loop for computing all CNFs of a given length, keep it fast!
;
;design:
;	for each complement
;	    compute new clause
;	    compute new truth-table
;	    compute new bits
;	    update stats
;
;group: truth.-table
; call: ztt4clause
; xref: zttcnf0

		public ztt4cnf0
		
ztt4cnf0:

;Original prolog
;  mov	ss:saveebp,ebp
;  mov	ebp,sp
;  NOTE regebp ;proccallx/retx can not be used since need ebp returned
;  ;proccallx <all>,<eax,ebx,ecx,edx,di,si>,'ztt4cnf0: gather statistics for truth table CNF\\Masks @eax.hex.ebx.hex.ecx.hex.edx.hex TT @di.hex.si.hex [varE=0800h, was0=0400h]'
;  test	ss:flagstt,ttlogcnf0
;  jz	aa@nolog
;  proclog <user>,<eax,ebx,ecx,edx,di,si>,'ztt4cnf0: Masks @eax.hex.ebx.hex.ecx.hex.edx.hex TT @di.hex.si.hex [varE=0800h, was0=0400h]'
;aa@nolog:
;  mov	ebp,ss:saveebp

		NOTE ;no PROC since using BP
  push	ebp ;CompMask
aa@loop:  ;
  push	ebp ;CompMask
  ifzero edx
  jz	aa@notedx
  jecxz	aa@notecx
  ifzero ebx
  jz	aa@notebx
  ifzero eax
  jz	aa@noteax
  not	eax 	;bit4 alternates every 8 ticks
  and	ebp,070000000h ;ttcompmask
  cmp	ebp,070000000h ;ttcompmask
  check ttcompmask,eq,0F0000000h    ;etc., see following
  je	aa@noteax
  not	eax ;7 of 8
aa@noteax:
  not	ebx
  and	ebp,030000000h ;ttcompmask
  cmp	ebp,030000000h ;ttcompmask
  je	aa@notebx 
  not	ebx ;3 of 4
aa@notebx: 
  test	ebp,010000000h ;bit6 of ttcompmask alternates every other tick
  jz	aa@keepecx
  not	ecx
aa@keepecx:
aa@notecx:
  not	edx 	;bit7 alternates every tick
aa@notedx:
  pop	ebp ;CompMask
  push	ebx
  or	ebx,eax
  or	ebx,ecx
  or	ebx,edx
  test	ebp,ttvarE
  jnz	aa@setvarE
  and	ebx,esi ;new truth table
  cmp	ebx,esi
  je	aa@same
  ifzero ebx
  jz	aa@nonsat ;returns to aa@sat
aa@sat:
  ztt4setbit edi,ebx ;eax,ecx,edx,esi,edi,>C
  jc	aa@next
  or	ebp,ttwas0
  jmp	aa@next
aa@same:  ;also from varE below
  inc	ebp
  check ttsamemask,eq,0ffh
aa@next:
  pop	ebx
  sub	ebp,(not ttcompmask) + 1
  check ttcompmask,eq,0F0000000h    ;only high bits effected
  jae	aa@loop 
  pop	edx ;CompMask
  and	edx,ttvarE+ttcompmask
  and	ebp,not (ttvarE+ttcompmask)
  or	ebp,edx
  NOTE regebp ;proccallx/retx can not be used since need ebp returned
  ret	;ret OK -- end of ztt4cnf0
  ;procretx <all>,<noresult,noflags>,'ztt4cnf0'

aa@setvarE:  ;EBX= new TT, ESI= old TT
  push	ebx
  and	ebx,esi ;0000FFFFh
  mov	bx,si 
  cmp	ebx,esi
  je	aa@same1
  ifzero ebx
  jz	aa@nonsatL
aa@satL:
  ztt4setbit edi,ebx ;eax,ecx,edx,esi,edi,>C
  jc	aa@var1
  or	ebp,ttwas0
  jmp	aa@var1
aa@same1:
  inc	ebp
  check ttsamemask,eq,0ffh
aa@var1:
  pop	ebx
  push	esi
  and	si,bx ;0FFFF0000h
  mov	ebx,esi
  pop	esi
  cmp	ebx,esi
  je	aa@same
  ifzero ebx
  jz	aa@nonsatH
aa@satH:
  ztt4setbit edi,ebx ;eax,ecx,edx,esi,edi,>C
  jc	aa@next
  or	ebp,ttwas0
  jmp	aa@next

aa@nonsat:
  push	edi
  mov	edi,[tt_global]
  mov	edi,[edi].curTask
  inc	[edi].totNonSat
  pop	edi
  jmp	aa@sat

aa@nonsatL:
  push	edi
  mov	edi,[tt_global]
  mov	edi,[edi].curTask
  inc	[edi].totNonSat
  pop	edi
  jmp	aa@satL

aa@nonsatH:
  push	edi
  mov	edi,[tt_global]
  mov	edi,[edi].curTask
  inc	[edi].totNonSat
  pop	edi
  jmp	aa@satH

  END
